/*
 * RCS $Id: redblack.h,v 1.9 2003/10/24 01:31:21 damo Exp $
 */

/*
   Redblack balanced tree algorithm
   Copyright (C) Damian Ivereigh 2000

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version. See the file COPYING for details.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Header file for redblack.c, should be included by any code that 
** uses redblack.c since it defines the functions 
*/ 
 
#ifndef RB_STATIC
#define RB_STATIC
#endif

#ifdef RUN
#include <stdio.h>
#include <stdlib.h>
#endif

typedef struct {
    int key;
    int value;
} my_data;

#define RB_CUSTOMIZE
#define RB_ENTRY(name)	rb##name
#define rbdata_t my_data

#define RB_CMP(s,t) 0
#define RB_LT(s,t) (s)->key < (t)->key
#define RB_GT(s,t) (s)->key > (t)->key

enum nodecolour { BLACK = 0, RED = 1};

#define RB_GET(x,y)		&x->y
#define RB_SET(x,y,v)		x->y = *(v)

struct RB_ENTRY(node)
{
	struct RB_ENTRY(node) *left;		/* Left down */
	struct RB_ENTRY(node) *right;		/* Right down */
	struct RB_ENTRY(node) *up;		/* Up */
	enum nodecolour colour;		/* Node colour */
	RB_ENTRY(data_t) key;		/* User's key (and data) */
};

struct RB_ENTRY(tree) {
  struct RB_ENTRY(node) *rb_root;
};

#define SIZE 1000
struct RB_ENTRY(node)* free_blocks[SIZE];

unsigned current_free = SIZE;

struct RB_ENTRY(node)* rb_alloc(void)
{
    struct RB_ENTRY(node) *retval;
    if (current_free == 0)
	return 0;
    retval = free_blocks[current_free - 1];
    current_free--;
    return retval;
}

struct RB_ENTRY(node) blocks[SIZE];
struct RB_ENTRY(tree) mytree;

/* Dummy (sentinel) node, so that we can make X->left->up = X
** We then use this instead of NULL to mean the top or bottom
** end of the rb tree. It is a black node.
**
** Initialization of the last field in this initializer is left implicit
** because it could be of any type.  We count on the compiler to zero it.
*/
struct RB_ENTRY(node) RB_ENTRY(_null);

/*={&RB_ENTRY(_null), &RB_ENTRY(_null), &RB_ENTRY(_null), BLACK};*/

#define RBNULL (&RB_ENTRY(_null))

#ifndef no_delete
RB_STATIC const RB_ENTRY(data_t) *RB_ENTRY(delete)(const RB_ENTRY(data_t) *, struct RB_ENTRY(tree) *);
#endif

#ifndef no_find
RB_STATIC const RB_ENTRY(data_t) *RB_ENTRY(find)(const RB_ENTRY(data_t) *, struct RB_ENTRY(tree) *);
#endif

#ifndef no_search
RB_STATIC const RB_ENTRY(data_t) *RB_ENTRY(search)(const RB_ENTRY(data_t) *, struct RB_ENTRY(tree) *);
#endif


/* Some useful macros */
#define rbmin(rbinfo) RB_ENTRY(lookup)(RB_LUFIRST, 0, (rbinfo))
#define rbmax(rbinfo) RB_ENTRY(lookup)(RB_LULAST, 0, (rbinfo))

#define assert(expr)


static void RB_ENTRY(_free)(struct RB_ENTRY(node) *x)
{
    free_blocks[current_free] = x;
    current_free++;
}

/* These functions are always needed */
static void RB_ENTRY(_left_rotate)(struct RB_ENTRY(node) **, struct RB_ENTRY(node) *);
static void RB_ENTRY(_right_rotate)(struct RB_ENTRY(node) **, struct RB_ENTRY(node) *);
static struct RB_ENTRY(node) *RB_ENTRY(_successor)(const struct RB_ENTRY(node) *);
static struct RB_ENTRY(node) *RB_ENTRY(_predecessor)(const struct RB_ENTRY(node) *);
static struct RB_ENTRY(node) *RB_ENTRY(_traverse)(int, const RB_ENTRY(data_t) * , struct RB_ENTRY(tree) *);

/* These functions may not be needed */

#ifndef no_delete
static void RB_ENTRY(_delete)(struct RB_ENTRY(node) **, struct RB_ENTRY(node) *);
static void RB_ENTRY(_delete_fix)(struct RB_ENTRY(node) **, struct RB_ENTRY(node) *);
#endif


/*
** OK here we go, the balanced tree stuff. The algorithm is the
** fairly standard red/black taken from "Introduction to Algorithms"
** by Cormen, Leiserson & Rivest. Maybe one of these days I will
** fully understand all this stuff.
**
** Basically a red/black balanced tree has the following properties:-
** 1) Every node is either red or black (colour is RED or BLACK)
** 2) A leaf (RBNULL pointer) is considered black
** 3) If a node is red then its children are black
** 4) Every path from a node to a leaf contains the same no
**    of black nodes
**
** 3) & 4) above guarantee that the longest path (alternating
** red and black nodes) is only twice as long as the shortest
** path (all black nodes). Thus the tree remains fairly balanced.
*/

#ifndef no_search
RB_STATIC const RB_ENTRY(data_t) *
RB_ENTRY(search)(const RB_ENTRY(data_t) *key, struct RB_ENTRY(tree) *rbinfo)
{
	struct RB_ENTRY(node) *x;

	if (rbinfo==0)
		return(0);

	x=RB_ENTRY(_traverse)(1, key, rbinfo);

	return((x==RBNULL) ? 0 : RB_GET(x, key));
}
#endif /* no_search */

#ifndef no_find
RB_STATIC const RB_ENTRY(data_t) * 
RB_ENTRY(find)(const RB_ENTRY(data_t) *key, struct RB_ENTRY(tree) *rbinfo)
{
	struct RB_ENTRY(node) *x;

	if (rbinfo==0)
		return(0);

	/* If we have a NULL root (empty tree) then just return */
	if (rbinfo->rb_root==RBNULL)
		return(0);

	x=RB_ENTRY(_traverse)(0, key, rbinfo);

	return((x==RBNULL) ? 0 : RB_GET(x, key));
}
#endif /* no_find */

#ifndef no_delete
RB_STATIC const RB_ENTRY(data_t) * 
RB_ENTRY(delete)(const RB_ENTRY(data_t) *key, struct RB_ENTRY(tree) *rbinfo)
{
	struct RB_ENTRY(node) *x;
	const RB_ENTRY(data_t) * y;

	if (rbinfo==0)
		return(0);

	x=RB_ENTRY(_traverse)(0, key, rbinfo);

	if (x==RBNULL)
	{
		return(0);
	}
	else
	{
		y=RB_GET(x, key);
		RB_ENTRY(_delete)(&rbinfo->rb_root, x);

		return(y);
	}
}
#endif /* no_delete */

/* --------------------------------------------------------------------- */

/* Search for and if not found and insert is true, will add a new
** node in. Returns a pointer to the new node, or the node found
*/
static struct RB_ENTRY(node) *
RB_ENTRY(_traverse)(int insert, const RB_ENTRY(data_t) *key, struct RB_ENTRY(tree) *rbinfo)
{
	struct RB_ENTRY(node) *x,*y,*z;
#ifdef RUN
	int level = 0;
#endif
	int found=0;

	y=RBNULL; /* points to the parent of x */
	x=rbinfo->rb_root;

#ifdef RUN
	/*printf("\nkey: %d, value: %d", *RB_GET(key,key), *RB_GET(key,value));*/
#endif
	/* walk x down the tree */
	while(x!=RBNULL && found==0)
	{
#ifdef RUN
	    level++;
#endif
		y=x;
		/* printf("key=%s, RB_GET(x, key)=%s\n", key, RB_GET(x, key)); */

		if (RB_LT(key, RB_GET(x,key)))
			x=x->left;
		else if (RB_GT(key, RB_GET(x,key)))
			x=x->right;
		else
			found=1;
	}

#ifdef RUN
	/*printf("\n  level: %d", level);*/
#endif


	if (found || !insert)
		return(x);

	z = rb_alloc();
	if (z==0)
	{
	    /* Whoops, no memory */
	    return(RBNULL);
	}

	RB_SET(z, key, key);
	z->up=y;
	if (y==RBNULL)
	{
		rbinfo->rb_root=z;
	}
	else
	{
		if (RB_LT(key, RB_GET(y,key)))
			y->left=z;
		else
			y->right=z;
	}

	z->left=RBNULL;
	z->right=RBNULL;

	/* colour this new node red */
	z->colour=RED;

	/* Having added a red node, we must now walk back up the tree balancing
	** it, by a series of rotations and changing of colours
	*/
	x=z;

	/* While we are not at the top and our parent node is red
	** N.B. Since the root node is garanteed black, then we
	** are also going to stop if we are the child of the root
	*/

	while(x != rbinfo->rb_root && (x->up->colour == RED))
	{
		/* if our parent is on the left side of our grandparent */
		if (x->up == x->up->up->left)
		{
			/* get the right side of our grandparent (uncle?) */
			y=x->up->up->right;
			if (y->colour == RED)
			{
				/* make our parent black */
				x->up->colour = BLACK;
				/* make our uncle black */
				y->colour = BLACK;
				/* make our grandparent red */
				x->up->up->colour = RED;

				/* now consider our grandparent */
				x=x->up->up;
			}
			else
			{
				/* if we are on the right side of our parent */
				if (x == x->up->right)
				{
					/* Move up to our parent */
					x=x->up;
					RB_ENTRY(_left_rotate)(&rbinfo->rb_root, x);
				}

				/* make our parent black */
				x->up->colour = BLACK;
				/* make our grandparent red */
				x->up->up->colour = RED;
				/* right rotate our grandparent */
				RB_ENTRY(_right_rotate)(&rbinfo->rb_root, x->up->up);
			}
		}
		else
		{
			/* everything here is the same as above, but
			** exchanging left for right
			*/

			y=x->up->up->left;
			if (y->colour == RED)
			{
				x->up->colour = BLACK;
				y->colour = BLACK;
				x->up->up->colour = RED;

				x=x->up->up;
			}
			else
			{
				if (x == x->up->left)
				{
					x=x->up;
					RB_ENTRY(_right_rotate)(&rbinfo->rb_root, x);
				}

				x->up->colour = BLACK;
				x->up->up->colour = RED;
				RB_ENTRY(_left_rotate)(&rbinfo->rb_root, x->up->up);
			}
		}
	}

	/* Set the root node black */
	(rbinfo->rb_root)->colour = BLACK;

	return(z);
}

/*
** Rotate our tree thus:-
**
**             X        rb_left_rotate(X)--->            Y
**           /   \                                     /   \
**          A     Y     <---rb_right_rotate(Y)        X     C
**              /   \                               /   \
**             B     C                             A     B
**
** N.B. This does not change the ordering.
**
** We assume that neither X or Y is NULL
*/

static void
RB_ENTRY(_left_rotate)(struct RB_ENTRY(node) **rootp, struct RB_ENTRY(node) *x)
{
	struct RB_ENTRY(node) *y;

	assert(x!=RBNULL);
	assert(x->right!=RBNULL);

	y=x->right; /* set Y */

	/* Turn Y's left subtree into X's right subtree (move B)*/
	x->right = y->left;

	/* If B is not null, set it's parent to be X */
	if (y->left != RBNULL)
		y->left->up = x;

	/* Set Y's parent to be what X's parent was */
	y->up = x->up;

	/* if X was the root */
	if (x->up == RBNULL)
	{
		*rootp=y;
	}
	else
	{
		/* Set X's parent's left or right pointer to be Y */
		if (x == x->up->left)
		{
			x->up->left=y;
		}
		else
		{
			x->up->right=y;
		}
	}

	/* Put X on Y's left */
	y->left=x;

	/* Set X's parent to be Y */
	x->up = y;
}

static void
RB_ENTRY(_right_rotate)(struct RB_ENTRY(node) **rootp, struct RB_ENTRY(node) *y)
{
	struct RB_ENTRY(node) *x;

	assert(y!=RBNULL);
	assert(y->left!=RBNULL);

	x=y->left; /* set X */

	/* Turn X's right subtree into Y's left subtree (move B) */
	y->left = x->right;

	/* If B is not null, set it's parent to be Y */
	if (x->right != RBNULL)
		x->right->up = y;

	/* Set X's parent to be what Y's parent was */
	x->up = y->up;

	/* if Y was the root */
	if (y->up == RBNULL)
	{
		*rootp=x;
	}
	else
	{
		/* Set Y's parent's left or right pointer to be X */
		if (y == y->up->left)
		{
			y->up->left=x;
		}
		else
		{
			y->up->right=x;
		}
	}

	/* Put Y on X's right */
	x->right=y;

	/* Set Y's parent to be X */
	y->up = x;
}

/* Return a pointer to the smallest key greater than x
*/
static struct RB_ENTRY(node) *
RB_ENTRY(_successor)(const struct RB_ENTRY(node) *x)
{
	struct RB_ENTRY(node) *y;

	if (x->right!=RBNULL)
	{
		/* If right is not NULL then go right one and
		** then keep going left until we find a node with
		** no left pointer.
		*/
		for (y=x->right; y->left!=RBNULL; y=y->left);
	}
	else
	{
		/* Go up the tree until we get to a node that is on the
		** left of its parent (or the root) and then return the
		** parent.
		*/
		y=x->up;
		while(y!=RBNULL && x==y->right)
		{
			x=y;
			y=y->up;
		}
	}
	return(y);
}

/* Return a pointer to the largest key smaller than x
*/
static struct RB_ENTRY(node) *
RB_ENTRY(_predecessor)(const struct RB_ENTRY(node) *x)
{
	struct RB_ENTRY(node) *y;

	if (x->left!=RBNULL)
	{
		/* If left is not NULL then go left one and
		** then keep going right until we find a node with
		** no right pointer.
		*/
		for (y=x->left; y->right!=RBNULL; y=y->right);
	}
	else
	{
		/* Go up the tree until we get to a node that is on the
		** right of its parent (or the root) and then return the
		** parent.
		*/
		y=x->up;
		while(y!=RBNULL && x==y->left)
		{
			x=y;
			y=y->up;
		}
	}
	return(y);
}

#ifndef no_delete
/* Delete the node z, and free up the space
*/
static void
RB_ENTRY(_delete)(struct RB_ENTRY(node) **rootp, struct RB_ENTRY(node) *z)
{
	struct RB_ENTRY(node) *x, *y;

	if (z->left == RBNULL || z->right == RBNULL)
		y=z;
	else
		y=RB_ENTRY(_successor)(z);

	if (y->left != RBNULL)
		x=y->left;
	else
		x=y->right;

	x->up = y->up;

	if (y->up == RBNULL)
	{
		*rootp=x;
	}
	else
	{
		if (y==y->up->left)
			y->up->left = x;
		else
			y->up->right = x;
	}

	if (y!=z)
	{
		RB_SET(z, key, RB_GET(y, key));
	}

	if (y->colour == BLACK)
		RB_ENTRY(_delete_fix)(rootp, x);

	RB_ENTRY(_free)(y);
}

/* Restore the reb-black properties after a delete */
static void
RB_ENTRY(_delete_fix)(struct RB_ENTRY(node) **rootp, struct RB_ENTRY(node) *x)
{
	struct RB_ENTRY(node) *w;

	while (x!=*rootp && x->colour==BLACK)
	{
		if (x==x->up->left)
		{
			w=x->up->right;
			if (w->colour==RED)
			{
				w->colour=BLACK;
				x->up->colour=RED;
				rb_left_rotate(rootp, x->up);
				w=x->up->right;
			}

			if (w->left->colour==BLACK && w->right->colour==BLACK)
			{
				w->colour=RED;
				x=x->up;
			}
			else
			{
				if (w->right->colour == BLACK)
				{
					w->left->colour=BLACK;
					w->colour=RED;
					RB_ENTRY(_right_rotate)(rootp, w);
					w=x->up->right;
				}


				w->colour=x->up->colour;
				x->up->colour = BLACK;
				w->right->colour = BLACK;
				RB_ENTRY(_left_rotate)(rootp, x->up);
				x=*rootp;
			}
		}
		else
		{
			w=x->up->left;
			if (w->colour==RED)
			{
				w->colour=BLACK;
				x->up->colour=RED;
				RB_ENTRY(_right_rotate)(rootp, x->up);
				w=x->up->left;
			}

			if (w->right->colour==BLACK && w->left->colour==BLACK)
			{
				w->colour=RED;
				x=x->up;
			}
			else
			{
				if (w->left->colour == BLACK)
				{
					w->right->colour=BLACK;
					w->colour=RED;
					RB_ENTRY(_left_rotate)(rootp, w);
					w=x->up->left;
				}

				w->colour=x->up->colour;
				x->up->colour = BLACK;
				w->left->colour = BLACK;
				RB_ENTRY(_right_rotate)(rootp, x->up);
				x=*rootp;
			}
		}
	}

	x->colour=BLACK;
}
#endif /* no_delete */


RB_ENTRY(data_t) x[1000]
= {
{2091565428,-911563765},
{-1817645480,1718559708},
{-2145591692,1116134493},
{-1527554227,-2036729790},
{651548850,-697956106},
{911096636,-427819588},
{-1211485523,256090520},
{2013675605,-1635362286},
{-1179140934,1043215027},
{-1816372436,-151448458},
{-1679473672,-2138871},
{-1818627580,-1606540838},
{-325551929,924285494},
{724524011,3005855},
{1963576003,-1174138769},
{-1214458382,1640041698},
{-1061459811,-1095418434},
{1150064164,397345432},
{-366966195,-1741598439},
{290869316,360802910},
{-1357668927,1853069812},
{-489709839,-1870718825},
{2115172043,-2135570156},
{1559391353,1428262445},
{-16417636,1203716116},
{1623064123,1848535989},
{2000545852,2028140446},
{495822762,-760352766},
{1526548113,-1788200980},
{-174530950,-174576500},
{261326455,1324366123},
{1790791214,146601677},
{-981342152,-826207003},
{803895693,-299469298},
{1969910124,-800889418},
{421585804,24257280},
{2027181115,-1816908223},
{-1176410769,196447762},
{-6664083,-1110971385},
{2125958806,1457180500},
{-460407207,1416739937},
{-663163199,-1921260798},
{1896335331,803838914},
{-2135908752,-2118571406},
{-493070807,-2069834636},
{-1631997056,-959735153},
{1639612636,1288632788},
{1180795952,-430421456},
{734233495,758022412},
{1415886981,-1882986370},
{1106466130,1335176153},
{-1367437831,-963013806},
{946776244,1382506568},
{-1004163545,2069950791},
{-117980964,2114380328},
{-1809429572,-792496428},
{96995883,1793576114},
{2082318402,530137881},
{2068420472,2136192118},
{1024094297,-1001399647},
{-1837087071,254676545},
{431180640,1190059502},
{-1388243620,1847938521},
{-1554274320,-1446368963},
{-2043470141,-1016669419},
{-474056617,872122131},
{-1264637687,-1683368098},
{-1812552878,-2112118878},
{773706664,48439621},
{396455464,1754048909},
{247971659,-687404980},
{1947455083,1306032534},
{-1078751725,-1674262823},
{1722025606,1947801785},
{-1259731907,-980427983},
{-2020970332,279473790},
{1164374989,1852080652},
{-1566412536,-1198045890},
{-765666047,1469146526},
{1871741167,261595410},
{-802913717,208333939},
{187275891,-1216635322},
{-1492338311,-224325288},
{-994642870,2013018343},
{831712491,-1572364666},
{1028777613,-2065025574},
{360894131,212597495},
{2014383864,-888160895},
{262567491,-1047269609},
{-1105146911,1896851182},
{166000940,-1815826570},
{-477491901,1172178023},
{-1251492965,-1559845771},
{-1219985949,-1927230592},
{-716592501,-1666031731},
{466537345,-1253306651},
{1503203813,1178948220},
{269513896,910868650},
{902008576,1596158647},
{113089190,-541221222},
{1217912063,-1497369086},
{1430971043,-1871294446},
{-1269460470,-583572614},
{-1796551508,-317619004},
{1989971878,1613138749},
{-277399330,-547516962},
{955735324,411424737},
{-617044984,285923021},
{906229305,-2125371413},
{-912924490,189796598},
{1091541092,1770899128},
{462764830,1892880625},
{1437601709,-417255186},
{1735059478,746925298},
{-960964110,-1612135668},
{-2110995995,1265482495},
{1533867149,1059458353},
{-945068823,-334209304},
{2022624164,629735220},
{-2046142917,643158406},
{1733667924,903778369},
{2025185740,-1864554908},
{661570055,269191650},
{-1079781969,-1494824055},
{630560757,765473463},
{-1656053712,1464755847},
{-1057307552,-1869099574},
{322139617,-144744151},
{1425187924,1229567647},
{-551742540,2141556219},
{-128362187,1838394998},
{-1797938366,-1674309627},
{1592534166,1856645055},
{-1391982210,484814810},
{1120924674,-1257772935},
{571205068,1351388184},
{-303555958,-1151799360},
{1380111090,-2068611095},
{-1409136153,569133786},
{-797774203,-1340606347},
{-388259208,1343928951},
{-76201362,2082047532},
{-2035174497,-492706680},
{447370257,-307805261},
{-130083400,1606224714},
{-457992905,-871184120},
{1139390959,779236283},
{1536277671,-32506235},
{1022126737,-1557294649},
{-1057695299,147970488},
{536711704,-1404282065},
{337288384,1735008177},
{-1024208299,-1977566275},
{-573617757,1201846616},
{-2141584653,1502690230},
{-1953441881,-1307431908},
{-764148003,1926413070},
{61617795,-619745802},
{-567577078,-1582411461},
{1588631545,1372237247},
{216201585,1947557401},
{-1903586621,355344410},
{835519620,1280582655},
{-1245747457,1469562825},
{-1157773824,-1684255114},
{1175392158,1817721119},
{33061487,500351051},
{-275101274,-2087049229},
{927424525,-1281122202},
{741559704,433731477},
{1347017605,71073530},
{-7361399,307693227},
{-641644955,-237043195},
{981962,-1412522701},
{-1159929598,1655736482},
{-1637978404,-1926333732},
{1334941397,-1101446780},
{548955250,-1506096778},
{1102008742,193770622},
{-970323016,-127021718},
{-345928542,1281521436},
{-271345580,380953954},
{311950594,718822403},
{-1484508851,220364632},
{1557644773,-600984054},
{709506022,-1855089231},
{-1536767492,773401259},
{-402840054,1971574831},
{-1264487890,1095679101},
{-1652173109,-819625001},
{1365839843,1766569054},
{2036119891,1374831646},
{1161261450,-1405397391},
{1935412884,-1534798590},
{-1014379179,-1684303589},
{-1061139178,18504182},
{-466523744,-1713976891},
{443332690,-1935336909},
{-1924565648,-364553931},
{-1393872353,-1111999259},
{1567717630,130933980},
{1482915977,532859947},
{-1826448771,-1767030742},
{1578074662,918247722},
{-752234559,-2015811612},
{-870859209,-857256134},
{-1210603540,-313659452},
{-813711994,864642400},
{-1795571713,1589592843},
{513825303,-1480815551},
{-44649031,-932894728},
{738413921,1519838452},
{188399767,538732081},
{-1044347641,-1707200185},
{-470314632,564459778},
{1299034700,1517972603},
{277194911,-2078126808},
{-935050940,983810642},
{364337708,-454940017},
{-1620364173,445916461},
{1215461237,720539269},
{233153109,-219792763},
{-153631796,1987494368},
{-65871072,-252305614},
{-170581473,-1134086198},
{1594407486,-2085332385},
{-1887118563,914013176},
{220083576,1021279380},
{-197094532,1547870044},
{2022074456,1133715383},
{-404431701,-1480889100},
{-184225778,-813368757},
{-1916143490,1460439665},
{1917868148,-1174328875},
{1791077157,870018503},
{-1300788042,1299221900},
{1485632885,-1928130216},
{319787456,1388446297},
{-1888088055,1194521931},
{307562816,279637437},
{1807861996,1439289987},
{990881857,1088265911},
{406993618,498152452},
{1702098649,1058180587},
{-349499001,-252987931},
{1753189107,-1393505615},
{-215802330,1418623701},
{300950939,-599587805},
{1290854593,1539575783},
{357529358,-768728168},
{-291158812,324820989},
{916771325,886021700},
{74345084,-1935556784},
{-632960212,1173844058},
{-1622122650,1372127689},
{2143245250,-1884402877},
{-1863109638,1420868997},
{-1756853059,837678325},
{-1308994138,1893785686},
{398017243,-1344552142},
{1771333370,-152412871},
{-166720382,1105572477},
{1771241041,-1564432974},
{154448364,906135190},
{-430363569,2090713351},
{221634056,-2087744852},
{1943407343,-1608910011},
{-318886981,-666303266},
{-1938603784,199546638},
{2144973067,1746683550},
{-1635335634,-1388249455},
{-317501433,1659858353},
{326421536,787597982},
{-236512437,2078683787},
{414422429,226377071},
{198605008,676046114},
{1711396475,2095049310},
{33228220,-1330132706},
{-710398441,1310864206},
{-717799510,1775764876},
{-278843320,-1592276639},
{1143738398,-547791889},
{-916103281,-143986590},
{1666849299,1104433433},
{-1695625456,-792164373},
{-598764659,-1249933884},
{545711945,-189579327},
{1236553353,-160248854},
{-233739767,-2030752430},
{1128518623,788155344},
{1205229633,1011846545},
{1752174638,-1955357410},
{-1020795425,-881686476},
{574851567,781239538},
{-397451154,-236591319},
{-788679421,1001917533},
{-1882018,252413705},
{1316246987,1027538898},
{1972710984,-1098588256},
{-2008181774,730234774},
{1506500154,-353569989},
{-1596057962,1054431585},
{-2105628392,1480029975},
{1687863078,-1769243354},
{1219650176,-672662091},
{-914742462,762406488},
{1592102785,169888125},
{-619973115,186810816},
{-364560501,-2060341343},
{-1675190332,1806643931},
{-110828538,232675594},
{-1473654119,-1340953065},
{1876092361,-934234163},
{-1566332498,-1383320973},
{-1410151747,-316268899},
{-305848350,2009114485},
{1043460520,483543667},
{-741777813,-57018633},
{-1281894210,-503814320},
{-144129528,-1153885329},
{1710898894,516089416},
{1637276495,2140699379},
{-1305135841,-782585809},
{812564346,-133909395},
{-418705179,279708423},
{760157014,2013089344},
{626421284,502632257},
{66703753,-643289054},
{1597153430,-759866743},
{-911305212,-1986890872},
{-1062567106,1679941597},
{1266209680,-1583672671},
{-283820407,-697914203},
{-52624503,-938544997},
{-932939915,-1449596600},
{950211193,629762668},
{1104780476,-1486827253},
{-2142711887,-2085043094},
{-1948227511,-785443149},
{-1283868388,-604384754},
{-2085013268,-464182569},
{1865818857,669199911},
{-696421173,-1650918793},
{2064198517,-2046513343},
{1522827226,-890097386},
{-564607778,-428367162},
{-297674228,1432305184},
{1435442484,1001747764},
{-1191505247,1363126228},
{-721174251,-1561115581},
{2003639002,-916744144},
{-1197748540,794365568},
{-54147139,-2091147583},
{-364636140,984106558},
{1585841409,-96369337},
{-158134384,99288702},
{416563786,-111996440},
{-1347257421,-407849814},
{590179100,976282998},
{2140367797,-213536111},
{1307627873,-665786206},
{-529231144,-187090734},
{-1918820859,841389150},
{1381577172,1279111228},
{255609134,-788395347},
{916736042,658795682},
{373355472,1147359134},
{-1258243599,413270911},
{-1518095292,2090651838},
{581466234,2014915341},
{-413718995,-282275042},
{1649498696,1445953477},
{805010290,1360490738},
{1283110338,-1081177826},
{-1506881453,2022705433},
{1441948558,1036029228},
{1367549813,712402990},
{-1109456342,-154335846},
{1288655527,-1170987856},
{-1997997158,548294757},
{-556322862,229786485},
{-1780588268,-1348415363},
{1392017624,-1757627477},
{-1512129844,1499042737},
{1625370476,533728410},
{643796763,2110518508},
{1124675144,575639217},
{-1603480402,1630712454},
{-885235822,1851752816},
{1023979603,-1658326515},
{128838032,-954996114},
{1914643402,225680239},
{-370517284,-1979873400},
{1447472905,-799282690},
{-929447329,-1399853055},
{-351372234,-397763699},
{-596097153,1754948499},
{-1154717399,1888590335},
{386778063,-777039661},
{1002658006,1546468047},
{1762128549,240005689},
{-1174282412,-1486782207},
{266726639,-341690176},
{-1622233475,1854706930},
{841486743,1690923736},
{580479501,1132275926},
{-2128133226,1121195382},
{-764551839,208999226},
{-1899037581,1441314546},
{-1488293295,882082439},
{1519764719,-225635159},
{1963226491,1065361527},
{1395030224,-505950081},
{-2040832040,795283641},
{-1400057947,857305554},
{289175157,396322879},
{-1515309695,966565436},
{470766883,463655628},
{924936844,-837562501},
{-622149230,1397909274},
{-364241997,-1845205531},
{1394601714,-131588359},
{-262000900,-664302152},
{-2055198693,591905147},
{750778808,-131660435},
{-1259634812,1064593646},
{-1284606851,-777767695},
{-1657596528,-1776451887},
{1932052676,-131945266},
{-1967179295,-1826346756},
{1252427538,-632872604},
{-1403790291,435683676},
{-635764687,-1228631143},
{-1162660174,1932487507},
{-2047897289,-59055811},
{-590964876,-1432365047},
{-1162197821,1554704648},
{1827843939,-1063590015},
{-1576733017,2092930413},
{-773756723,840125112},
{948474552,1157437262},
{-67013749,263434456},
{1505212676,-1525437560},
{-10514319,324227810},
{227145624,-7121961},
{965016566,-840293745},
{619690948,-1227305596},
{1166870951,-655485433},
{-1988703316,2078996581},
{607880169,-2023144456},
{517437658,-2061786750},
{63933679,1700664117},
{2117463370,-198585219},
{631014407,-1329262224},
{1824619474,-496669378},
{-329031233,-811868632},
{817593859,78484319},
{938328956,1391083225},
{1355243496,1983718591},
{1543917240,-1710256480},
{-1097201491,301409604},
{800237016,831437605},
{-2039756001,1321379620},
{332778146,-391554487},
{-729146170,428362385},
{-424183935,-975758092},
{-977348291,-1051120959},
{-201408448,-1887678846},
{375251843,-879342189},
{1672276631,1652861701},
{-891403129,-1785852570},
{-812522278,227069633},
{499962216,-1061659927},
{-327393981,556089935},
{701927407,-1678524156},
{73621038,-54462146},
{-1460324157,1624361740},
{-497100481,402182476},
{1956497467,-465201237},
{70474807,15720897},
{1347702889,1494583175},
{967315196,-832355495},
{-1510400357,-425094810},
{1782115357,2032894276},
{1413361291,2080073006},
{492028513,1151481734},
{-375323232,922578429},
{-161470194,-1016235064},
{-2015680151,-1036786305},
{266524054,-357270654},
{85054155,-1213462158},
{1261922420,-814641731},
{461161506,-290778187},
{-1082002924,-1254232352},
{-911459245,-1903124197},
{-967626460,434618874},
{-1135210204,1671334663},
{1193063332,-1031175021},
{-663487134,-997803352},
{1196026317,-305725970},
{412192526,462894195},
{-2038764079,804037959},
{1746453680,-1875669115},
{-1331554121,1402419457},
{1634209521,-1721014936},
{807392142,-1957906692},
{-701859841,-1248805184},
{473187833,-1407590334},
{-3551664,232092161},
{-1440127428,-697571772},
{1595808629,-1605693954},
{-588561581,1099159781},
{-1786900235,755076659},
{1277397379,-1917688007},
{-1373726101,1729659732},
{426057049,584810645},
{712767300,324440810},
{1853032927,-1044508541},
{-1447769822,2081237012},
{-1363348610,2111618858},
{1705604214,1449573928},
{-1284711574,-667291519},
{-91304182,-210384313},
{1239234021,-946969945},
{-1612097764,1957412402},
{-145939690,471521305},
{-1357156143,-1482557268},
{-1438187814,1575786303},
{-18420701,-1831826326},
{-2097654870,1814086302},
{-601101800,558836636},
{-672181925,-1644853495},
{-629733689,-1454443085},
{161735897,221899243},
{183755411,-2091913145},
{-1003665343,470767958},
{250608736,-177734671},
{-506744834,242255363},
{-1868807940,-800698808},
{159414933,530302258},
{1752850905,1564822001},
{1763226086,1285995194},
{1343694108,-1421441075},
{-603156845,-1221961847},
{2047319670,1054545540},
{1244033843,-1514427490},
{-2097892903,1358864352},
{-465711307,898972810},
{-1726428307,-1985665367},
{-2103032408,1141668595},
{-207117957,1499118672},
{-160351408,1664040384},
{1973293390,-466768286},
{-1867903180,2035201721},
{-1354780,-1928136624},
{-988265388,-538144651},
{995936950,1909380793},
{1500503284,-103244931},
{-1319191763,370412396},
{28476826,1539730602},
{-458071408,-924130154},
{-1840211018,1996803377},
{-2058430652,-161148237},
{309598492,-1466422440},
{-1075619259,-109537505},
{2131885326,-2136260136},
{1815146039,635652965},
{246158016,-1697909434},
{2034398809,-882720366},
{-1286404167,1421208329},
{362012800,-2096563565},
{-134165806,-1977146736},
{-496338295,-1638423960},
{-151983030,-1784333414},
{1986389333,1663268287},
{300402652,-789293127},
{-1688250014,1914443992},
{-256569940,2024521706},
{-623041591,-1290194033},
{-1241620332,-1966992463},
{-1059057715,-71214952},
{-937649576,1030686166},
{2128051099,-861618964},
{-1760164880,468287916},
{1175441936,2128101780},
{1423416933,2050594616},
{1831257813,656182273},
{165678475,515869477},
{1933623864,288293287},
{295149149,1500345054},
{1865450260,556686703},
{-1865392461,-631579090},
{-1260222879,177878977},
{469853413,-435131040},
{-1274294948,-1079022077},
{1567004402,517225966},
{-34492336,-2076934410},
{-2124498707,-605358085},
{1644276187,-1173113163},
{39535759,108052102},
{269405829,1835914802},
{366921026,892886821},
{1214875508,933705749},
{-1419316805,-1987540880},
{108967321,1783450627},
{1199303811,-1928447700},
{-1403369853,2058281352},
{611285821,595180603},
{-304780886,1619950893},
{407173952,165434238},
{189779317,108962074},
{402588055,915421615},
{1866010948,597537757},
{-792713659,1613927463},
{-1332579024,1491505476},
{-1543574973,677072848},
{-222814189,-372574124},
{-514894160,-104553953},
{-1637148351,166285951},
{162535333,-403808135},
{1633096161,798893745},
{1379847701,-326926036},
{1790062727,-361324275},
{1757633944,-887026265},
{-1187937217,-462952626},
{-1797608752,-260021206},
{-781022606,-1179526214},
{-1508014724,-824620572},
{437196593,1271953409},
{-1888851878,234503723},
{-1269827309,-570101363},
{1634225123,712942090},
{569244369,230026422},
{1743401295,-1506642929},
{-916354786,1320498192},
{-2122019844,-1270513099},
{1621149435,-81293435},
{-1993862305,746464975},
{911625158,-1996183572},
{626210318,1073045325},
{-1737536731,1414824162},
{-1793119376,1710394632},
{1312668939,-838876376},
{-881044293,1698937925},
{356888687,1717227746},
{-2096028519,-1319473310},
{1632775412,1258536662},
{1729815249,1306956770},
{1487170087,-505839062},
{1348280728,-1694796329},
{-952159223,1800671154},
{-348094884,-40429949},
{-984873144,-908250915},
{2073866768,1793588045},
{-1248413812,-1670660615},
{-1088739399,-3130097},
{1404431022,1561231654},
{577719206,-1118694121},
{-1494733393,1215598298},
{-2124007059,-400328201},
{-866202294,1894814570},
{-1388819695,-204260299},
{298076379,1320495273},
{557571003,-1815241617},
{1832594530,-419746479},
{190050312,-1792708518},
{162838580,-1380128335},
{-2116851415,-1100758546},
{941552006,-1628155476},
{-241524788,-672637912},
{-1559853041,1926416945},
{-1169274908,1931942422},
{-675680891,-1117215879},
{1742645854,403031260},
{1623263918,2044846425},
{-1416132485,-280718099},
{-1846210791,-393088896},
{1593805702,-479853098},
{911376346,976115492},
{1852961311,787553396},
{964779519,-1246889757},
{-519726378,317942511},
{1285596965,1588138334},
{1213426611,-764446645},
{-1630246361,852405686},
{-75778702,2082743511},
{-1591045181,920322654},
{1671949621,-1677681398},
{-1465161678,1049187218},
{279524876,1980394705},
{-1582435018,2119367644},
{1142422177,-1560531737},
{-1893866183,-1701753307},
{1597712560,-809316285},
{-400904878,-190313289},
{-1290031342,414616344},
{-497598546,-445113781},
{683449645,-194761617},
{-1699074229,-766778997},
{902976255,-1319197650},
{-445280862,471520028},
{362996831,-936783868},
{-862991605,-689272561},
{-847608197,-613248672},
{-420208820,-2122330166},
{-1336387291,1222059481},
{336253749,1581990631},
{-21181917,-1888588795},
{1627171942,361972553},
{-1493461490,-606586835},
{278369931,214508747},
{-252264654,323701223},
{1886119872,-1776131940},
{-468155877,1963001387},
{-692062603,2043639631},
{3729356,-1179707299},
{-1997369954,1473586425},
{692952626,-1552317244},
{-1228549009,-2097300685},
{-447099762,5803489},
{-572066003,1260759501},
{485159554,1844631507},
{326301963,1890466678},
{1378048620,-1960932619},
{316886535,-423799075},
{-1219725008,-1981161719},
{1595458662,1781604032},
{57013369,80810762},
{-1421859303,1669301253},
{270363856,-936757368},
{959752954,926697139},
{1602430728,-1418252853},
{-2027181506,-781672777},
{1670684600,-1400213236},
{-398845956,1013734391},
{674257495,1084938698},
{-288631769,-1855652008},
{1973205415,-1155661445},
{1324666127,-975792648},
{-675021047,-551446823},
{2088613782,-1654141772},
{-1948470448,582990702},
{-1941204982,-655692082},
{-543786599,-416611979},
{199903436,1391413994},
{-769706068,-811099925},
{-334466728,-726501755},
{2033454427,-1564206680},
{-463584530,1438915648},
{1977736166,464404086},
{-1471147796,-91952681},
{713243302,699869116},
{-633345000,-938771942},
{343391254,1349939616},
{-1389280793,1687345757},
{1206906301,-1102352486},
{27449947,-1935062161},
{268207189,-1629159965},
{1372078769,-1606805732},
{-82193658,-850985161},
{102990664,2118589728},
{1392145324,-1124239901},
{1250003420,-1909194042},
{1768170800,-1979916905},
{1257847366,-1653323212},
{-2019478018,-1117925892},
{-708628527,1181911562},
{736439178,1506360845},
{-1675170180,1285811760},
{-651028973,-1553774325},
{-926670024,62974819},
{32063951,195494276},
{1662903494,-668482225},
{836333370,-631487356},
{-216172447,-188300435},
{1698811995,-1792041467},
{-1024644557,-2146330383},
{1192937593,-1429222276},
{-720622132,-519481145},
{-1387605171,2042597401},
{-416309262,290439235},
{-202520449,650120881},
{-587841862,-1371304521},
{385598949,-1365668736},
{-1946231919,938433390},
{769652884,963809833},
{-1712369114,-95477451},
{-50782752,1855515655},
{554187184,-1719734646},
{920397834,1175400529},
{1060793197,1647409906},
{1415083742,-709152005},
{-82626818,58034230},
{-1867757414,195071070},
{-357651394,-237649869},
{-281139726,875526161},
{2091636767,-1592034543},
{670149040,-225307002},
{1019172826,1029850735},
{-590828254,1500604934},
{1239104488,1525734774},
{432999579,244537486},
{2011354366,1199085121},
{427869233,980337355},
{374904794,1236573306},
{1044839675,-628512423},
{-1187113568,203301774},
{1249882252,303913247},
{-1115062224,-1300881347},
{-1402970218,191229129},
{2027606189,1411851380},
{-1317441885,-680014127},
{-669593338,-589916866},
{-1230984093,-1630961866},
{729308800,2048555835},
{1390686700,173210497},
{2034633456,1906327080},
{-872027029,-409477314},
{1340386812,-1060141345},
{2066033782,55208080},
{1509259257,155092837},
{-1863554236,863373715},
{-1472834696,277086793},
{969431337,-1074303025},
{-1924389134,-2146701171},
{-783219018,1401082086},
{82539884,486467231},
{2053895256,589233490},
{-1186800258,2008430568},
{382897382,1348250235},
{1488766539,424286179},
{2075634318,1167295374},
{-184252013,-353988671},
{1575207735,-943593214},
{-2068810542,1658837637},
{-1971735757,1243409368},
{-844913806,1913923214},
{-1828676925,936059619},
{592543181,-1480674044},
{711451354,-225750007},
{-1666998668,-1863174398},
{187325789,-604986726},
{-2089344410,-2083874833},
{216928194,777165239},
{136673195,142169540},
{-1658294607,-987826288},
{291091309,-624047992},
{821502455,-1898246030},
{-1429536679,1435740230},
{125737097,-1725515151},
{9584023,-1981194379},
{694223242,-1288354217},
{-1154480776,648762909},
{-361269269,768645877},
{-1957605882,-1654903242},
{95008221,972642164},
{872118757,-993923720},
{-154142054,-1065453714},
{-86377912,390238294},
{-1510040416,1611497592},
{-961632620,-1182329280},
{-1749366051,-1975262468},
{1196185379,43629326},
{740864038,668254853},
{-1277170236,463355003},
{-86650149,-1595038367},
{-280489515,-953629215},
{-1516385524,160083827},
{530420013,1653436386},
{1145904617,1691790701},
{-816413620,1485799949},
{604207869,-1022776806},
{-1668749727,1165206747},
{867456879,1928427937},
{-450554479,-790992677},
{1152216664,-149990851},
{-146005964,1814939093},
{1137158795,-1486516930},
{-947182525,1363366426},
{831404385,-1285284851},
{1724791394,-2108070341},
{-915769416,-1592711019},
{476104488,1185780250},
{1187296552,-1163474182},
{-991714691,-477729451},
{-42692126,-1934962254},
{641595965,393781398},
{-1389504617,-1825670815},
{-1982074200,1411576988},
{-1527788492,2116683864},
{1093024407,-1484730484},
{343267920,1466139828},
{-1331650646,1316372451},
{-1760632013,216733775},
{-107212442,-1727744680},
{142097775,-316045774},
{1044705709,-727048295},
{-436943473,281419885},
{-1851228978,1662805191},
{119387740,1381995862},
{-777847003,677897510},
{-1654735723,-645739698},
{-1505146459,1111181576},
{1631932579,1623932561},
{1007366464,-954165193},
{899745720,826794169},
{854251799,-650504300},
{-1323577084,277194277},
{-435901180,-344935808},
{-1849723073,1705101420},
{1216676238,-707146470},
{-2096026698,1230279809},
{387092591,-252735803},
{1272546432,-1890329361},
{1421645685,-952954123},
{278366529,-551679561},
{37756540,-657420965},
{-1887290877,-947381849},
{-883838086,1657608636},
{-1838123702,-1788290651},
{777366151,-1364570957},
{973072779,2029227166},
{-1500766684,-844973937},
{-2064846042,345917275},
{145785221,1842598547},
{1218376977,-842075917},
{-149541419,-787112552},
{-1092605702,-2065397766},
{-321462361,1772400788},
{242418098,1541769956},
{341358300,1374834348},
{-1635913745,213839628},
{-576606215,1667291416},
{-551451275,1749714608},
{714084590,541174167},
{-817263067,1529178497},
{-1741573361,-285776921},
{-660756295,-1579692593},
{-2066807123,1260835289},
{-380311268,-1189452538},
{1383629444,-604417028},
{1071143991,-488589765},
{595991500,-1444435225},
{1488852218,-404496347},
{1980225711,-919449742},
{-370336498,-1495612954},
{1158701981,-1844897194},
{2083639513,348447085},
{-1596126391,1107382381},
{-1159512101,-157531668},
{-478252431,-1021274520},
{-280613320,-902562932},
{2116044136,606853742},
{-26187714,-628556959},
{-563947645,576419296},
{-1212836590,-1819888820},
{1877730494,-1181266067},
{-228276053,497778900},
{793280981,-526149486},
{797340640,1187249684},
{-1103241445,-1104798114},
{-1749086476,-166835206},
{2059325579,-665653983},
{1668120884,1786157854},
{-1117846710,-140643030},
{11254361,2052651247},
{-9991703,-1878429643},
{-1860029504,-1573233471},
{-564815458,1945216460},
{116419362,-331968759},
{-707822520,1152939814},
{1335681899,-1626827487},
{-1611554294,-1652167724},
{-258408799,279860349},
{1988652409,-1075732845},
{929002023,369501918},
{1032030605,1660235768},
{131802748,1176711261},
{842343623,1068085025},
{973781448,336586835},
{-167415733,2060660749},
{1398606234,-694225035},
{203843794,793489612},
{123956901,479179652},
{1330226260,156953826},
{1984136300,-154170629},
{1515832516,709538149},
{1769684518,-779982411},
{887395895,215441135},
{886196963,-1378293477},
{532140403,75402927},
{-1187290972,-1812084810},
{-546514352,1832062593},
{1155220077,-1062163280},
{-1061005228,865699261},
{670342766,-472822701},
{-191357537,1696257143},
{-1597930007,20111609},
{-951366415,-638684066},
{-650097116,361207324}};

int start()
{
    int i, k;

    mytree.rb_root = RBNULL;
    RBNULL->left = RBNULL;
    RBNULL->right = RBNULL;
    RBNULL->up = RBNULL;
    RBNULL->colour = BLACK;

    for (i = 0; i < 1000; ++i) {
	rbsearch(&(x[i]), &mytree);
    }

    rbdelete(&x[500], &mytree);
    rbfind(&x[999], &mytree);
}
